import base64

from typing import Sequence
import kvix
from kvix import ActionType
from kvix.impl import BaseActionType, BasePlugin
from kvix.plugin.builtin.machinist import BaseMachinist


class Action(BaseMachinist):
    def _get_text(self, query: str) -> str:
        return base64.b64decode(self.action_type.context.ui.paste_from_clipboard()).decode("UTF-8")


class Plugin(BasePlugin):
    def _create_single_action_type(self) -> ActionType:
        return BaseActionType(
            self.context,
            "base64-decode",
            "Base64 decode",
            action_factory=Action,
        )

    def get_actions(self) -> Sequence[kvix.Action]:
        return [Action(self._single_action_type, "Base64 decode", "decode base64")]



import sys
from argparse import ArgumentParser
from enum import Enum
from typing import cast, Sequence

import pynput

import kvix.impl
import kvix.ui.tk
import kvix.ui.tray
from kvix import Action, ActionRegistry, Context, Item
from kvix.conf import Conf, StorConf
from kvix.impl import BaseActionRegistry, BaseItem, BaseItemAlt, BaseItemSource
from kvix.l10n import _
from kvix.plugin import Plugin as PanPlugin
from kvix.remote import Client as RemoteClient
from kvix.remote import Server
from kvix.stor import YamlFile
from kvix.util import (
    apply_template,
    ensure_type,
    get_cache_dir,
    get_config_dir,
    get_data_dir,
)

activate_action_text = _("Activate").setup(ru_RU="Выпуолнить", de_DE="Aktivieren")
edit_action_text = _("Edit Action: {{action_title}} ({{action_type_title}})").setup(
    ru_RU='Редактировать действие "{{action_title}}" ({{action_type_title}})',
    de_DE='Aktion Bearbeiten: "{{action_title}}" ({{action_type_title}})',
)
delete_action_text = _("Remove Action: {{action_title}} ({{action_type_title}})").setup(
    ru_RU='Удалить действие "{{action_title}}" ({{action_type_title}})',
    de_DE='Aktion Löschen "{{action_title}}" ({{action_type_title}})',
)

ui_scope_title_text = _("UI")
ui_tk_scope_title_text = _("TK")


def load_conf():
    result = StorConf(YamlFile(get_config_dir().joinpath("config.yaml")))
    result.load()
    return result


class App(Context):
    def __init__(self, conf: Conf | None = None):
        self._conf = cast(Conf, conf)
        self._conf.scope("ui", "UI")

    def run(self):
        self.init_conf()
        self.init_action_stor()
        self.init_cache_stor()
        self.init_action_registry()
        self.load_actions()
        self.init_tray()

    def init_conf(self):
        if not self._conf:
            self._conf = self.create_conf()

    def create_conf(self) -> Conf:
        return load_conf()

    def init_action_stor(self):
        self.action_stor = self.create_action_stor()
        self.action_stor.data = []

    def create_action_stor(self):
        return YamlFile(get_data_dir().joinpath("actions.yaml"))

    def init_cache_stor(self):
        self._cache_stor = self.create_cache_stor()
        self._cache_stor.data = self._cache_stor.data or {}

    def create_cache_stor(self):
        return YamlFile(get_cache_dir().joinpath("cache.yaml"))

    def init_action_registry(self):
        self._action_registry = self.create_action_registry()
        self._action_registry.action_types

    def create_action_registry(self) -> ActionRegistry:
        return BaseActionRegistry(self.action_stor)

    def load_actions(self):
        # load action types from plugins
        pan_plugin = PanPlugin(self)
        for action_type in pan_plugin.get_action_types():
            self._action_registry.add_action_type(action_type)

        # load action from config
        self.action_registry.load()

        # load known action type ids
        self._cache_stor.load()
        self._cache_stor.data = self._cache_stor.data or {}
        cache = ensure_type(self._cache_stor.data, dict)
        known_action_types_ids = set(ensure_type(cache.get("known_action_type_ids", []), list))

        # from plugins load actions which are of unknown action types
        for action in pan_plugin.get_actions():
            if not action.action_type.id in known_action_types_ids:
                self.action_registry.actions.append(action)
        self.action_registry.save()

        # save known action type id cache
        cache["known_action_type_ids"] = list(
            set([id for id in self._action_registry.action_types])
        )
        self._cache_stor.save()

    def init_tray(self):
        self.tray = kvix.ui.tray.TrayIcon()
        self.tray.on_show = self.activate_action_selector
        self.tray.on_quit = self.quit
        self.tray.run(self.init_ui)

    def init_ui(self):
        self._ui = kvix.ui.tk.Ui(self.conf)
        self._ui.on_ready(self._on_ui_ready)
        self.ui.run()

    def _on_ui_ready(self) -> None:
        self.init_action_selector()
        self.register_global_hotkeys()

    def init_action_selector(self):
        self.action_selector = self.ui.selector()
        self.action_selector.title = "kvix!!!"

        def edit_action(action: Action) -> None:
            dialog = action.action_type.context.ui.dialog(action.action_type.create_editor)
            dialog.value = action

            def on_ok():
                self.action_registry.save()

            dialog.on_ok = on_ok
            dialog.activate()

        def delete_action(action: Action) -> None:
            self.action_registry.actions.remove(action)
            self.action_registry.save()
            self.action_selector.activate()

        def search(query: str) -> Sequence[Item]:
            result: list[Item] = []
            for action in self.action_registry.actions:
                for item in action.search(query):
                    alts = list(item.alts)

                    def edit_this_action(action: Action = action):
                        edit_action(action)

                    alts.append(
                        BaseItemAlt(
                            apply_template(
                                str(edit_action_text),
                                action_title=action.title,
                                action_type_title=action.action_type.title,
                            ),
                            edit_this_action,
                        )
                    )

                    def delete_this_action(action: Action = action):
                        delete_action(action)

                    alts.append(
                        BaseItemAlt(
                            apply_template(
                                str(delete_action_text),
                                action_title=action.title,
                                action_type_title=action.action_type.title,
                            ),
                            delete_this_action,
                        )
                    )
                    result.append(BaseItem(str(item), alts))
            return result

        self.action_selector.item_source = BaseItemSource(search)

    def activate_action_selector(self):
        self.action_selector.activate()

    def quit(self):
        self.conf.save()
        self.action_registry.save()
        self.ui.destroy()
        self.tray.stop()

        # self.server.stop()

    def register_global_hotkeys(self):
        activate_window_hotkey = self.conf.item("activate_window_hotkey").setup(
            title="Activate Window Hotkey",
            default="<ctrl>+;",
            read_mapping=str,
        )
        pynput.keyboard.GlobalHotKeys(
            {activate_window_hotkey.read(): self.activate_action_selector}
        ).start()


class RemoteCommand(Enum):
    ping = "ping"
    activate = "activate"
    quit = "quit"

    def __str__(self):
        return self.value


def run(*args: str):
    parser = ArgumentParser(prog="kvix")
    parser.add_argument(
        "-r",
        "--remote",
        type=RemoteCommand,
        choices=list(RemoteCommand),
        help="activate remote command",
    )
    # parser.add_argument('-a', '--activate', action='store_true', help='show action selector immediately after start')
    # parser.add_argument('-f', '--force', action='store_true', help='skip existing running instance check')
    parsed_args = parser.parse_args(args)

    conf = load_conf()
    remote_client = RemoteClient(conf)
    if parsed_args.remote:
        if RemoteCommand.ping == parsed_args.remote:
            remote_client.ping()
        elif RemoteCommand.activate == parsed_args.remote:
            remote_client.activate()
        elif RemoteCommand.quit == parsed_args.remote:
            remote_client.quit()
    elif not remote_client.is_server_ready():
        app = App(conf)
        server = Server(app, app.activate_action_selector)
        try:
            server.start()
            app.run()
        finally:
            server.stop()
    else:
        print("already running")


def main():
    run(*(sys.argv[1:]))


if __name__ == "__main__":
    main()



from typing import Any

import pynput
from typing import cast, Callable, Sequence

from kvix import Action, ActionType, Context, DialogBuilder, ItemAlt
from kvix.impl import (
    BaseAction,
    BaseActionType,
    BasePlugin,
    BaseItemAlt,
)
from kvix.util import apply_template

from kvix.l10n import _
from kvix.util import query_match

text_text = _("Text").setup(ru_RU="Текст")
type_text = _("Type text").setup(ru_RU="Печатать текст", de_DE="Text eingeben")
copy_text = _("Copy to clipboard").setup(
    ru_RU="Копировать в буфер обмена", de_DE="In die Zwischenablage kopieren"
)
paste_text = _("Copy&Paste").setup(
    ru_RU="Копировать&Вставить",
    de_DE="In die Zwischenablage kopieren&einfügen",
)


class MachinistActionType(BaseActionType):
    def __init__(self, context: Context):
        BaseActionType.__init__(self, context, "machinist", "Machinist")

    def create_default_action(
        self,
        title: str,
        description: str = "",
        pattern: str = "",
        **config: Any,
    ) -> Action:
        return Machinist(self, "", title, description or "", **config)

    def action_from_config(self, value: Any):
        self._assert_config_valid(value)
        return Machinist(self, value["text"], value.get("title"), value.get("description"))

    def create_editor(self, builder: DialogBuilder) -> None:
        builder.create_entry("text", str(text_text))
        super().create_editor(builder)

        def load(value: Any | None):
            if isinstance(value, Machinist):
                builder.widget("text").set_value(value.text)

        builder.on_load(load)

        def save(value: Any | None = {}) -> Any:
            if isinstance(value, Machinist):
                value.text = builder.widget("text").get_value()
            else:
                value = Machinist(
                    self.context.action_registry.action_types["machinist"],
                    builder.widget("text").get_value(),
                    builder.widget("title").get_value(),
                    builder.widget("description").get_value(),
                )
            return value

        builder.on_save(save)


class BaseMachinist(BaseAction):
    def _create_single_item_alts(self, query: str) -> Sequence[ItemAlt]:
        return [
            self._create_single_item_type_alt(lambda: self._get_text(query)),
            self._create_single_item_copy_alt(lambda: self._get_text(query)),
            self._create_single_item_paste_alt(lambda: self._get_text(query)),
        ]

    def _create_single_item_type_alt(self, action: Callable[[], str]) -> ItemAlt:
        return BaseItemAlt(type_text, lambda: self._type_text(action()))

    def _create_single_item_copy_alt(self, action: Callable[[], str]) -> ItemAlt:
        return BaseItemAlt(copy_text, lambda: self._copy_text(action()))

    def _create_single_item_paste_alt(self, action: Callable[[], str]) -> ItemAlt:
        return BaseItemAlt(paste_text, lambda: self._paste_text(action()))

    def _get_text(self, query: str) -> str:
        raise NotImplementedError()

    def _type_text(self, text: str):
        self.action_type.context.ui.hide()
        pynput.keyboard.Controller().type(text)

    def _copy_text(self, text: str):
        self.action_type.context.ui.hide()
        self.action_type.context.ui.copy_to_clipboard(text.encode())

    def _paste_text(self, text: str):
        self.action_type.context.ui.hide()
        old_clipboard_content = None
        try:
            old_clipboard_content = self.action_type.context.ui.paste_from_clipboard()
        except Exception as e:
            print(e)
        self.action_type.context.ui.copy_to_clipboard(text.encode())
        from pynput.keyboard import Key, Controller

        keyboard = Controller()
        keyboard.press(cast(str, Key.ctrl.value))
        keyboard.press("v")
        keyboard.release("v")
        keyboard.release(cast(str, Key.ctrl.value))
        if old_clipboard_content is not None:
            try:
                self.action_type.context.ui.copy_to_clipboard(old_clipboard_content)
            except Exception as e:
                print("error copying to clipboard", e)


class Machinist(BaseMachinist):
    def __init__(
        self,
        action_type: ActionType,
        text: str,
        title: str | None = None,
        description: str = "",
    ):
        title = title or 'type text "' + text + '"'
        BaseAction.__init__(self, action_type, title, description or title)
        self.text = text

    def _get_text(self, query: str) -> str:
        return apply_template(self.text, {"query": query})

    def _match(self, query: str) -> bool:
        if query_match(query, self.text):
            return True
        return BaseAction._match(self, query)

    def to_config(self):
        result = BaseAction.to_config(self)
        result["text"] = self.text
        return result


class Plugin(BasePlugin):
    def _create_single_action_type(self) -> ActionType:
        return MachinistActionType(self.context)



from typing import Sequence
import kvix
from kvix import ActionType
from kvix.impl import BaseActionType, BasePlugin
from kvix.plugin.builtin.machinist import BaseMachinist
from kvix.l10n import _

import uuid

action_title_text = _("Generate Random UUID").setup(ru_RU="Генерировать случайный UUID")


class Action(BaseMachinist):
    def _get_text(self, query: str) -> str:
        return str(uuid.uuid4())


class Plugin(BasePlugin):
    def _create_single_action_type(self) -> ActionType:
        return BaseActionType(
            self.context,
            "kvix.random-uuid",
            str(action_title_text),
            action_factory=Action,
        )

    def get_actions(self) -> Sequence[kvix.Action]:
        return [
            Action(
                self._single_action_type,
                str(action_title_text),
                "\n".join(action_title_text.values()),
            )
        ]



import PIL.Image
import pystray
import pkg_resources
import threading
from typing import Callable


class TrayIcon:
    def __init__(self):
        self.on_show: Callable[[], None] = lambda: None
        self.on_quit: Callable[[], None] = lambda: None

    def run(self, callback: Callable[[], None]):
        image = PIL.Image.open(pkg_resources.resource_filename("kvix", "logo.jpg"))
        menu = pystray.Menu(
            pystray.MenuItem("kvix", self.on_show, default=True),
            pystray.MenuItem("exit", self.on_quit),
        )
        self._icon = pystray.Icon("kvix", image, "kvix", menu, visible=True)
        if "darwin" in pystray.Icon.__module__:
            self._icon.run_detached(None)
        else:
            threading.Thread(target=self._icon.run).start()
        callback()

    def stop(self):
        self._icon.stop()



from kvix import ActionType, Context
from kvix.impl import BaseAction, BaseActionType, BasePlugin
from kvix.l10n import _
import subprocess
import shlex
from typing import Any
from kvix.util import apply_template

import os

action_type_title_text = _("Run Command").setup(ru_RU="Запуск команды")
command_title_text = _("Command").setup(ru_RU="Команда")
default_action_title_text = action_type_title_text
default_action_title_description = " ".join(default_action_title_text.values())


class Action(BaseAction):
    def _run(self, **args: Any) -> None:
        self.action_type.context.ui.hide()
        env = os.environ.copy()
        env["KVIX_CLIPBOARD"] = ""
        clipboard_bytes = b""
        try:
            clipboard_bytes = self.action_type.context.ui.paste_from_clipboard()
            env["KVIX_CLIPBOARD"] = str(clipboard_bytes.decode("UTF-8"))
        except Exception as e:
            print("error pasting from clipboard", e)
        shell_command = apply_template(str(self._config["command"]), env)
        shell_args = shlex.split(shell_command)
        subprocess.Popen(shell_args, env=env)


class Plugin(BasePlugin):
    def __init__(self, context: Context):
        super().__init__(context)

    def _create_single_action_type(self) -> ActionType:
        return BaseActionType(
            self.context,
            "kvix.plugin.builtin.shell",
            str(action_type_title_text),
            Action,
            {"command": command_title_text},
        )



from kvix import ActionType, Context, Item
from kvix.impl import (
    BaseAction,
    BaseActionType,
    BasePlugin,
    BaseItem,
    BaseItemAlt,
    execute_text,
)
from kvix.l10n import _
import kvix
import webbrowser
import urllib.parse as urlparse
from typing import Sequence

action_type_title_text = _("Google Search").setup(ru_RU="Поиск в google")
search_google_text = _('Google for "{query}"').setup(ru_RU='Загуглить "{query}"')
search_duck_text = _('DuckDuckGo for "{query}"').setup(ru_RU='Найти в DuckDuckGo "{query}"')

command_title_text = _("Command").setup(ru_RU="Команда")
default_action_title_text = action_type_title_text
default_action_title_description = " ".join(default_action_title_text.values())


class Action(BaseAction):
    def _search_google(self, query: str) -> None:
        self.action_type.context.ui.hide()
        webbrowser.open("https://www.google.com/search?btnI=1&q=" + urlparse.quote(query))

    def _search_duck(self, query: str) -> None:
        self.action_type.context.ui.hide()
        webbrowser.open("https://duckduckgo.com/?q=" + urlparse.quote("\\ " + query))

    def search(self, query: str) -> Sequence[Item]:
        if not query:
            return []
        if 4 > len(query):
            return []
        return [
            BaseItem(
                str(search_google_text).format(query=query),
                [BaseItemAlt(str(execute_text), lambda: self._search_google(query))],
            ),
            BaseItem(
                str(search_duck_text).format(query=query),
                [BaseItemAlt(str(execute_text), lambda: self._search_duck(query))],
            ),
        ]


class Plugin(BasePlugin):
    def __init__(self, context: Context):
        super().__init__(context)

    def _create_single_action_type(self) -> ActionType:
        return BaseActionType(
            self.context,
            "kvix.plugin.builtin.websearch",
            str(action_type_title_text),
            Action,
        )

    def get_actions(self) -> Sequence[kvix.Action]:
        return [Action(self._single_action_type, str(action_type_title_text))]
